from abconfig.common import Dict
from abconfig.utils import GetAttrs


class Settings:
    __hidesettings__ = True
    __hidesettings_exclude__ = []
    __file__ = False
    __file_required__ = False
    __env__ = True
    __prefix__ = None
    __vault__ = False
    __vault_required__ = True
    __settings__ = (
        '__hidesettings__',
        '__hidesettings_exclude__',
        '__file__',
        '__file_required__',
        '__env__',
        '__prefix__',
        '__vault__',
        '__vault_required__'
    )


class HideSettings(Dict):
    def __init__(self, obj: Dict):
        super().__init__(self.remove_settings(obj))

    def remove_settings(self, obj: Dict) -> [tuple]:
        result = []
        exclude = obj.get('__hidesettings_exclude__', [])
        if obj.get('__hidesettings__', True) is True:
            for k,v in dict(obj).items():
                if not k in Settings.__settings__ or k in exclude:
                    result.append((k,v))
        return result


import warnings
from functools import wraps

from abconfig.common import Dict


def ignore_warnings(f):
    @wraps(f)
    def inner(*args, **kwargs):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("ignore")
            response = f(*args, **kwargs)
        return response
    return inner


class GetAttrs(Dict):
    def __init__(self, obj=None, settings=None):
        if not settings: settings = []
        super().__init__({
            str(k): v for k,v in self.is_type(obj).__dict__.items()
            if k[:1] != '_' or k in settings
        })


class Close(Dict):
    def __init__(self, obj: Dict):
        super().__init__(Dict(obj).fmap(self.set_default_type))

    @staticmethod
    def set_default_type(k, v):
        if Dict.is_dict(v):
            return (k, Close(v))
        elif Dict.is_list(v):
            return (k, Dict.is_type(v)([None if isinstance(i, type) else i for i in v]))
        else:
            return (k, None if isinstance(v, type) else v)


class Switch(type):
    def __call__(cls, obj: Dict) -> Dict:
        if cls.__target__.enabled(obj) is True:
            return obj.bind(cls.__target__)
        else:
            return obj



try:
    from hvac import Client as VaultClient
    from hvac.exceptions import VaultError
    from hvac.exceptions import Unauthorized as VaultUnauthorized
except ImportError:
    pass

try:
    import boto3
except ImportError:
    pass

from abconfig.common import Dict
from abconfig.env import OSEnviron

from abconfig.utils import Switch, Close, ignore_warnings


class VaultData(OSEnviron):
    _config_scheme = Dict(
        auth_type='token',
        data_type='json',
        addr=str,
        token=str,
        path=str,
        kv_version=2,
        header_value=str,
        role=str
    )

    def __init__(self, obj: Dict):
        self._required = obj.get('__vault_required__', True)
        self._config = (self._config_scheme + obj.get('__vault__')).bind(Close)
        try:
            self._cache = self._request()
        except VaultError as error:
            if self._required is True:
                raise VaultError(error)
            self._cache = {}

        if self._config['data_type'] == 'kv':
            super().__init__(obj + self.read(obj, obj.get('__prefix__')))
        elif self._config['data_type'] == 'json':
            super().__init__(obj + self._cache)
        else:
            raise ValueError(f'only supported "kv" or "json"')

    def env(self, prefix: str, k: str, v: any) -> tuple:
        key = k if k != '__vault__' else 'vault'
        if self.is_dict(v):
            return (key, self.read(v, self.concat(prefix,key)))
        else:
            return (key, self._cache.get(self.concat(prefix, key).upper(), None))

    def _request(self) -> Dict:
        if self._config['kv_version'] == 2:
            return Dict(self._auth.v2.read_secret_version(
                path=self._config['path']
            )['data']['data'])
        elif self._config['kv_version'] == 1:
            return Dict(self._auth.v1.read_secret(
                path=self._config['path']
            )['data'])
        else:
            return Dict()

    @property
    @ignore_warnings
    def _auth(self):
        client = VaultClient(url=self._config['addr'])
        auth = self._config['auth_type']
        token = self._config['token']

        if token and auth == 'token':
            client.token = token
        elif auth == 'aws_iam':
            session = boto3.Session()
            creds = session.get_credentials()
            kwargs = [
                self._config['header_value'],
                self._config['role']
            ]

            client.auth_aws_iam(
                creds.access_key,
                creds.secret_key,
                creds.token,
                **{
                    k:v for k,v in self._config.items()
                    if v and k in kwargs
                }
            )

        if client.is_authenticated() is False:
            raise VaultUnauthorized(f'auth_type: {auth}')

        return client.secrets.kv

    @staticmethod
    def enabled(obj: Dict):
        return Dict.is_dict(obj.get('__vault__'))


class Vault(metaclass=Switch):
    __target__ = VaultData



import json

try:
    import yaml
except ImportError:
    pass

try:
    import toml
except ImportError:
    pass

from os.path import splitext
from configparser import ConfigParser

from abconfig.common import Dict


class Reader(Dict):
    def __init__(self, obj: Dict):
        self._path = obj.get('__file__')
        if self._path:
            prefix = obj.get('__prefix__', None)
            if prefix:
                super().__init__(obj + self._read[prefix])
            else:
                super().__init__(obj + self._read)
        else:
            super().__init__(obj)

    @property
    def _read(self) -> dict:
        try:
            with open(self._path, 'r') as fd:
                result = self._driver(fd)
                if not self.is_dict(result):
                    raise IOError
                return result
        except Exception:
            return self.__mempty__

    def _driver(self, fd) -> dict:
        raise NotImplementedError


class Yaml(Reader):
    __extensions__ = ('yml', 'yaml')

    def _driver(self, fd) -> dict:
        return yaml.load(fd, Loader=yaml.FullLoader)


class Json(Reader):
    __extensions__ = ('json',)

    def _driver(self, fd) -> dict:
        return json.load(fd)


class Toml(Reader):
    __extensions__ = ('toml',)

    def _driver(self, fd) -> dict:
        return toml.load(fd)


class Ini(Reader):
    __extensions__ = ('ini',)

    @property
    def _read(self) -> dict:
        driver = ConfigParser()
        driver.read(self._path)
        return driver._sections


class Format(type):
    def __call__(cls, obj: Dict) -> Dict:
        path = obj.get('__file__')
        if path:
            return obj.do(*cls._format(path, cls.__formats__))
        else:
            return obj

    @staticmethod
    def _format(path, formats) -> tuple:
        _, extension = splitext(path)
        for format_ in formats:
            for e in format_.__extensions__:
                if e == extension[1:]:
                    return (format_,)
        return formats


class File(metaclass=Format):
    __formats__ = (Json, Yaml, Toml, Ini)



from collections import UserDict, OrderedDict


class Type:
    __types__ = (
        (int, float, str, bool),
        (list, tuple, set, frozenset)
    )

    @staticmethod
    def is_type(item: any) -> type:
        return item if isinstance(item, type) else type(item)

    @staticmethod
    def is_list(item: any) -> bool:
        if isinstance(item, Type.__types__[1]):
            return True
        else:
            return False


class Item(Type):
    """ Converting an item type to type from a related subset of types. """
    __mempty__ = None

    def __init__(self, item: any):
        if item != self.__mempty__:
            self._type = self.is_type(item)
            self._set = [s for s in self.__types__ if self._type in s][0]
        self.data = item

    def __add__(self, operand: any) -> any:
        """ Converting operand to type(self). """
        if self.data == self.__mempty__:
            return operand
        elif operand == self.__mempty__ or isinstance(operand, type):
            return self.data
        elif self.is_list(self.data) and self.data:
            return self._type(
                [
                    self._type_by_index(index)(i)
                    if not isinstance(i, type) else i
                    for index,i in enumerate(list(operand))
                ]
            )
        elif not self.is_type(operand) in self._set:
            raise TypeError
        else:
            return self._type(operand)

    def _type_by_index(self, index: int) -> type:
        if index >= len(self.data):
            return self.is_type(list(self.data)[0])
        return self.is_type(list(self.data)[index])


class Dict(Type, UserDict):
    """ Wrapper to build a dictionary processing chain.

    (+) - atomic update values ​​of the first dictionary
    to values ​​from second dictionary with support recursive processing
    nested dictionaries.
    """

    __mempty__ = dict()

    def fmap(self, f: type) -> dict:
        """ Applies function to each pair of key values ​​in dict. """
        return Dict([f(k,v) for k,v in self.items()])

    def bind(self, f: type) -> dict:
        """ Applies function to object. """
        return f(self)

    def do(self, *args: [type]) -> dict:
        acc = args[0](self)
        for obj in args[1:]:
            acc = acc.bind(obj)
        return acc

    def __add__(self, operand: dict) -> dict:
        if operand == self.__mempty__:
            return self.data
        elif self.data == self.__mempty__:
            return operand

        return self.fmap(lambda k,v:
            (k, Dict(v) + operand.get(k, self.__mempty__))
            if self.is_dict(v) else
            (k, Item(v) + operand.get(k, Item.__mempty__))
        )

    @staticmethod
    def is_dict(obj: any) -> bool:
        return isinstance(obj, (dict, Dict, OrderedDict, UserDict))



from os import environ

from abconfig.common import Dict
from abconfig.utils import Switch


class OSEnviron(Dict):
    def __init__(self, obj: Dict):
        self.__list_separator__ = ','
        super().__init__(obj + self.read(obj, obj.get('__prefix__', None)))

    def read(self, obj: Dict, prefix: str) -> Dict:
        return Dict([self.env(prefix,k,v) for k,v in obj.items()])

    def env(self, prefix: str, k: str, v: any) -> tuple:
        if self.is_dict(v):
            return (k, self.read(v, self.concat(prefix,k)))

        var = environ.get(self.concat(prefix, k).upper(), None)
        if not var: return (k,v)
        if self.is_list(v):
            return (k, self.is_type(v)(var.split(self.__list_separator__)))
        else:
            return (k, self.is_type(v)(var))

    @staticmethod
    def enabled(obj: Dict):
        return obj.get('__env__', True)

    @staticmethod
    def concat(*args: [str], **kwargs):
        s = kwargs.get('space', '_')
        return s.join(filter(lambda x: True if x else False, args))


class Environment(metaclass=Switch):
    __target__ = OSEnviron



import re
import pathlib

from setuptools import setup
from sys import version_info


if version_info < (3, 6, 0):
    raise RuntimeError("abconfig requires Python 3.6+")

here = pathlib.Path(__file__).parent
txt = (here / 'abconfig' / '__init__.py').read_text('utf-8')

try:
    version = re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
except IndexError:
    raise RuntimeError('Unable to determine version.')

with open('README.md', 'r') as f:
    long_description = f.read()

args = dict(
    name='abconfig',
    version=version,
    author='Alexander Shevchenko',
    author_email='kudato@me.com',
    description='Simple and powerful configuration manager for Python projects',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kudato/abconfig.git",
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    packages=['abconfig'],
    extras_require={
        'yaml': ['pyyaml>=5.1'],
        'toml': ['toml>=0.10.0'],
        'vault': ['hvac>=0.9.6, <0.11.2'],
        'aws': ['boto3>=1.10.45']
    },
    python_requires='>=3.6.0'
)

setup(**args)


__version__ = '1.0.10'

from abconfig.common import Dict

from abconfig.file import File
from abconfig.env import Environment
from abconfig.vault import Vault

from abconfig.utils import GetAttrs, Close
from abconfig.settings import Settings, HideSettings


class ABConfig(Dict, Settings):
    __pipeline__ = (File, Environment, Vault)

    def __init__(self, obj=None):
        super().__init__(
            GetAttrs(
                obj if obj else self,
                settings=self.__settings__
            )
            .do(*self.__pipeline__)
            .do(HideSettings, Close)
            .items()
        )
        self.__dict__.update(self)
